import os
from typing import List, Optional, Tuple, Union

import torch
import torch.utils.checkpoint
from memoria import Abstractor, Memoria
from torch import nn
from torch.nn import BCEWithLogitsLoss, CrossEntropyLoss, MSELoss
from transformers import (
    AutoConfig,
    AutoModel,
    AutoModelForSequenceClassification,
    BertForSequenceClassification,
    BertModel,
    ElectraForSequenceClassification,
    ElectraModel,
)
from transformers.modeling_outputs import (
    BaseModelOutputWithPastAndCrossAttentions,
    BaseModelOutputWithPoolingAndCrossAttentions,
    SequenceClassifierOutput,
)
from transformers.modeling_utils import ModuleUtilsMixin, PreTrainedModel
from transformers.models.bert.modeling_bert import BertAttention, BertEmbeddings, BertIntermediate, BertPooler
from transformers.pytorch_utils import apply_chunking_to_forward
from transformers.utils import logging

from .configuration_memoria_bert import MemoriaBertConfig

logger = logging.get_logger(__name__)


class MemoriaBertOutput(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.dense = nn.Linear(config.intermediate_size, config.hidden_size)
        self.LayerNorm = nn.LayerNorm(config.hidden_size, eps=config.layer_norm_eps)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)

    def forward(self, hidden_states: torch.Tensor, input_tensor: torch.Tensor) -> torch.Tensor:
        hidden_states = self.dense(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)
        return hidden_states


class MemoriaBertLayer(nn.Module):
    def __init__(self, config: MemoriaBertConfig, use_memory_attention: bool = False):
        super().__init__()
        self.chunk_size_feed_forward = config.chunk_size_feed_forward
        self.seq_len_dim = 1
        self.attention = BertAttention(config)
        self.is_decoder = config.is_decoder
        self.use_memory_attention = use_memory_attention
        if self.use_memory_attention:
            self.memory_attention = BertAttention(config)
        self.intermediate = BertIntermediate(config)
        self.output = MemoriaBertOutput(config)

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        past_key_value: Optional[Tuple[Tuple[torch.FloatTensor]]] = None,
        working_memory_states: Optional[torch.FloatTensor] = None,
        working_memory_attention_mask: Optional[torch.FloatTensor] = None,
        memory_states: Optional[torch.FloatTensor] = None,
        memory_attention_mask: Optional[torch.FloatTensor] = None,
        output_attentions: Optional[bool] = False,
    ) -> Tuple[torch.Tensor]:
        # decoder uni-directional self-attention cached key/values tuple is at positions 1,2
        self_attn_past_key_value = past_key_value[:2] if past_key_value is not None else None
        self_attention_outputs = self.attention(
            hidden_states,
            attention_mask,
            head_mask,
            output_attentions=output_attentions,
            past_key_value=self_attn_past_key_value,
        )
        attention_output = self_attention_outputs[0]

        outputs = self_attention_outputs[1:]  # add self attentions if we output attention weights

        if working_memory_states is not None and working_memory_attention_mask is not None:
            if not hasattr(self, "memory_attention"):
                raise ValueError(
                    f"If `working_memory_states` are passed, {self} has to be instantiated with memory-attention layers"
                    " by setting `use_memory_attention=True`"
                )

            # cross_attn cached key/values tuple is at positions 3,4 of past_key_value tuple
            attn_past_key_value = None
            working_memory_attention_outputs = self.memory_attention(
                attention_output,
                attention_mask,
                head_mask,
                working_memory_states,
                working_memory_attention_mask,
                attn_past_key_value,
                output_attentions,
            )
            attention_output = working_memory_attention_outputs[0]
            outputs = (
                outputs + working_memory_attention_outputs[1:]
            )  # add cross attentions if we output attention weights

        if memory_states is not None and memory_attention_mask is not None:
            if not hasattr(self, "memory_attention"):
                raise ValueError(
                    f"If `memory_states` are passed, {self} has to be instantiated with memory-attention layers"
                    " by setting `use_memory_attention=True`"
                )

            # cross_attn cached key/values tuple is at positions 3,4 of past_key_value tuple
            attn_past_key_value = None
            memory_attention_outputs = self.memory_attention(
                attention_output,
                attention_mask,
                head_mask,
                memory_states,
                memory_attention_mask,
                attn_past_key_value,
                output_attentions,
            )
            attention_output = memory_attention_outputs[0]
            outputs = outputs + memory_attention_outputs[1:]  # add cross attentions if we output attention weights

        layer_output = apply_chunking_to_forward(
            self.feed_forward_chunk, self.chunk_size_feed_forward, self.seq_len_dim, attention_output
        )
        outputs = (layer_output,) + outputs

        return outputs

    def feed_forward_chunk(self, attention_output):
        intermediate_output = self.intermediate(attention_output)
        layer_output = self.output(intermediate_output, attention_output)
        return layer_output


class MemoriaBertEncoder(nn.Module):
    def __init__(self, config: MemoriaBertConfig):
        super().__init__()
        self.config = config
        self.layer = nn.ModuleList(
            [
                MemoriaBertLayer(config, use_memory_attention=i > config.memory_layer_index)
                for i in range(config.num_hidden_layers)
            ]
        )
        self.memoria = Memoria(
            num_reminded_stm=config.memoria_num_reminded_stm,
            stm_capacity=config.memoria_stm_capacity,
            ltm_search_depth=config.memoria_ltm_search_depth,
            initial_lifespan=config.memoria_initial_lifespan,
            num_final_ltms=config.memoria_num_reminded_ltm,
            device=config.memoria_device,
        )
        self.abstractor = Abstractor(
            num_memories=config.memoria_num_memories,
            hidden_dim=config.hidden_size,
            feedforward_dim=config.intermediate_size,
        )
        self.gradient_checkpointing = False

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: Optional[torch.FloatTensor] = None,
        head_mask: Optional[torch.FloatTensor] = None,
        encoder_hidden_states: Optional[torch.FloatTensor] = None,
        encoder_attention_mask: Optional[torch.FloatTensor] = None,
        past_key_values: Optional[Tuple[Tuple[torch.FloatTensor]]] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = False,
        output_hidden_states: Optional[bool] = False,
        return_dict: Optional[bool] = True,
    ) -> Union[Tuple[torch.Tensor], BaseModelOutputWithPastAndCrossAttentions]:
        device = hidden_states.device
        batch_size = hidden_states.size(0)
        all_hidden_states = () if output_hidden_states else None
        all_self_attentions = ()
        all_cross_attentions = ()
        all_working_memory_attentions = ()
        all_memory_attentions = ()

        next_decoder_cache = () if use_cache else None
        working_memory_states = None
        working_memory_attention_mask = None
        memory_states = None
        memory_indices = None
        memory_attention_mask = None
        for i, layer_module in enumerate(self.layer):
            if output_hidden_states:
                all_hidden_states = all_hidden_states + (hidden_states,)

            layer_head_mask = head_mask[i] if head_mask is not None else None
            past_key_value = past_key_values[i] if past_key_values is not None else None

            if self.gradient_checkpointing and self.training:
                if use_cache:
                    logger.warning(
                        "`use_cache=True` is incompatible with gradient checkpointing. Setting `use_cache=False`..."
                    )
                    use_cache = False

                def create_custom_forward(module):
                    def custom_forward(*inputs):
                        return module(*inputs, past_key_value, True)

                    return custom_forward

                layer_outputs = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(layer_module),
                    hidden_states,
                    attention_mask,
                    layer_head_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    working_memory_states,
                    working_memory_attention_mask,
                    memory_states,
                    memory_attention_mask,
                )
            else:
                layer_outputs = layer_module(
                    hidden_states,
                    attention_mask,
                    layer_head_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    past_key_value,
                    working_memory_states,
                    working_memory_attention_mask,
                    memory_states,
                    memory_attention_mask,
                    True,
                )

            hidden_states = layer_outputs[0]
            if use_cache:
                next_decoder_cache += (layer_outputs[-1],)
            all_self_attentions = all_self_attentions + (layer_outputs[1],)
            if working_memory_states is not None:
                all_working_memory_attentions = all_working_memory_attentions + (layer_outputs[2],)
                all_memory_attentions = all_memory_attentions + (layer_outputs[3],)

            if i == self.config.memory_layer_index:
                working_memory_states = self.abstractor(hidden_states)
                working_memory_attention_mask = torch.ones([batch_size, working_memory_states.size(1)], device=device)
                working_memory_attention_mask = ModuleUtilsMixin.get_extended_attention_mask(
                    self, working_memory_attention_mask, None, dtype=hidden_states.dtype
                )

                self.memoria.add_working_memory(working_memory_states)

                # Remind with working memory added at previous timestep
                memory_states, memory_indices = self.memoria.remind()
                memory_attention_mask = torch.ones_like(memory_indices, device=device)
                memory_attention_mask.masked_fill_(memory_indices.to(device) < 0, 0)
                memory_attention_mask = ModuleUtilsMixin.get_extended_attention_mask(
                    self, memory_attention_mask, None, dtype=hidden_states.dtype
                )

        if output_hidden_states:
            all_hidden_states = all_hidden_states + (hidden_states,)

        if working_memory_states is not None:
            memory_weights = torch.cat(all_memory_attentions, dim=1)
            memory_weights = memory_weights.view(
                memory_weights.size(0), memory_weights.size(1) * memory_weights.size(2), memory_weights.size(3)
            ).mean(dim=1)
            memory_weights *= memory_weights.size(1) * self.config.memoria_lifespan_extend_scale
            self.memoria.adjust_lifespan_and_memories(memory_indices, memory_weights)

        if not output_attentions:
            all_self_attentions = None
            all_cross_attentions = None

        if not return_dict:
            return tuple(
                v
                for v in [
                    hidden_states,
                    next_decoder_cache,
                    all_hidden_states,
                    all_self_attentions,
                    all_cross_attentions,
                ]
                if v is not None
            )
        return BaseModelOutputWithPastAndCrossAttentions(
            last_hidden_state=hidden_states,
            past_key_values=next_decoder_cache,
            hidden_states=all_hidden_states,
            attentions=all_self_attentions,
            cross_attentions=all_cross_attentions,
        )


class MemoriaBertPreTrainedModel(PreTrainedModel):
    """
    An abstract class to handle weights initialization and a simple interface for downloading and loading pretrained
    models.
    """

    config_class = MemoriaBertConfig
    base_model_prefix = "memoria_bert"
    supports_gradient_checkpointing = True
    _keys_to_ignore_on_load_missing = [r"position_ids"]

    def _init_weights(self, module):
        """Initialize the weights"""
        if isinstance(module, nn.Linear):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def _set_gradient_checkpointing(self, module, value=False):
        if isinstance(module, MemoriaBertEncoder):
            module.gradient_checkpointing = value

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: Optional[Union[str, os.PathLike]], *model_args, **kwargs):
        config = AutoConfig.from_pretrained(pretrained_model_name_or_path, **kwargs)
        if config.model_type == "memoria_bert":
            return super().from_pretrained(pretrained_model_name_or_path, *model_args, **kwargs)
        elif config.model_type == "bert":
            return cls.from_pretrained_bert(pretrained_model_name_or_path, *model_args, **kwargs)
        elif config.model_type == "electra":
            return cls.from_pretrained_electra(pretrained_model_name_or_path, *model_args, **kwargs)
        else:
            raise ValueError("Unsupported model type")

    @classmethod
    def from_pretrained_bert(cls, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def from_pretrained_electra(cls, *args, **kwargs):
        raise NotImplementedError


class MemoriaBertModel(MemoriaBertPreTrainedModel):
    """

    The model can behave as an encoder (with only self-attention) as well as a decoder, in which case a layer of
    cross-attention is added between the self-attention layers, following the architecture described in [Attention is
    all you need](https://arxiv.org/abs/1706.03762) by Ashish Vaswani, Noam Shazeer, Niki Parmar, Jakob Uszkoreit,
    Llion Jones, Aidan N. Gomez, Lukasz Kaiser and Illia Polosukhin.

    To behave as an decoder the model needs to be initialized with the `is_decoder` argument of the configuration set
    to `True`. To be used in a Seq2Seq model, the model needs to initialized with both `is_decoder` argument and
    `add_cross_attention` set to `True`; an `encoder_hidden_states` is then expected as an input to the forward pass.
    """

    def __init__(self, config: MemoriaBertConfig, add_pooling_layer=True):
        super().__init__(config)
        self.config = config

        self.embeddings = BertEmbeddings(config)
        self.encoder = MemoriaBertEncoder(config)

        self.pooler = BertPooler(config) if add_pooling_layer else None

        # Initialize weights and apply final processing
        self.post_init()

    def get_input_embeddings(self):
        return self.embeddings.word_embeddings

    def set_input_embeddings(self, value):
        self.embeddings.word_embeddings = value

    def _prune_heads(self, heads_to_prune):
        """
        Prunes heads of the model. heads_to_prune: dict of {layer_num: list of heads to prune in this layer} See base
        class PreTrainedModel
        """
        for layer, heads in heads_to_prune.items():
            self.encoder.layer[layer].attention.prune_heads(heads)

    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple[torch.Tensor], BaseModelOutputWithPoolingAndCrossAttentions]:
        r"""
        encoder_hidden_states  (`torch.FloatTensor` of shape `(batch_size, sequence_length, hidden_size)`, *optional*):
            Sequence of hidden-states at the output of the last layer of the encoder. Used in the cross-attention if
            the model is configured as a decoder.
        encoder_attention_mask (`torch.FloatTensor` of shape `(batch_size, sequence_length)`, *optional*):
            Mask to avoid performing attention on the padding token indices of the encoder input. This mask is used in
            the cross-attention if the model is configured as a decoder. Mask values selected in `[0, 1]`:

            - 1 for tokens that are **not masked**,
            - 0 for tokens that are **masked**.
        past_key_values (`tuple(tuple(torch.FloatTensor))` of length `config.n_layers` with each tuple having 4 tensors of shape `(batch_size, num_heads, sequence_length - 1, embed_size_per_head)`):
            Contains precomputed key and value hidden states of the attention blocks. Can be used to speed up decoding.

            If `past_key_values` are used, the user can optionally input only the last `decoder_input_ids` (those that
            don't have their past key value states given to this model) of shape `(batch_size, 1)` instead of all
            `decoder_input_ids` of shape `(batch_size, sequence_length)`.
        use_cache (`bool`, *optional*):
            If set to `True`, `past_key_values` key value states are returned and can be used to speed up decoding (see
            `past_key_values`).
        """
        output_attentions = output_attentions if output_attentions is not None else self.config.output_attentions
        output_hidden_states = (
            output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        )
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        if self.config.is_decoder:
            use_cache = use_cache if use_cache is not None else self.config.use_cache
        else:
            use_cache = False

        if input_ids is not None and inputs_embeds is not None:
            raise ValueError("You cannot specify both input_ids and inputs_embeds at the same time")
        elif input_ids is not None:
            input_shape = input_ids.size()
        elif inputs_embeds is not None:
            input_shape = inputs_embeds.size()[:-1]
        else:
            raise ValueError("You have to specify either input_ids or inputs_embeds")

        batch_size, seq_length = input_shape
        device = input_ids.device if input_ids is not None else inputs_embeds.device

        # past_key_values_length
        past_key_values_length = past_key_values[0][0].shape[2] if past_key_values is not None else 0

        if attention_mask is None:
            attention_mask = torch.ones(((batch_size, seq_length + past_key_values_length)), device=device)

        if token_type_ids is None:
            if hasattr(self.embeddings, "token_type_ids"):
                buffered_token_type_ids = self.embeddings.token_type_ids[:, :seq_length]
                buffered_token_type_ids_expanded = buffered_token_type_ids.expand(batch_size, seq_length)
                token_type_ids = buffered_token_type_ids_expanded
            else:
                token_type_ids = torch.zeros(input_shape, dtype=torch.long, device=device)

        # We can provide a self-attention mask of dimensions [batch_size, from_seq_length, to_seq_length]
        # ourselves in which case we just need to make it broadcastable to all heads.
        extended_attention_mask: torch.Tensor = self.get_extended_attention_mask(attention_mask, input_shape)

        # If a 2D or 3D attention mask is provided for the cross-attention
        # we need to make broadcastable to [batch_size, num_heads, seq_length, seq_length]
        if self.config.is_decoder and encoder_hidden_states is not None:
            encoder_batch_size, encoder_sequence_length, _ = encoder_hidden_states.size()
            encoder_hidden_shape = (encoder_batch_size, encoder_sequence_length)
            if encoder_attention_mask is None:
                encoder_attention_mask = torch.ones(encoder_hidden_shape, device=device)
            encoder_extended_attention_mask = self.invert_attention_mask(encoder_attention_mask)
        else:
            encoder_extended_attention_mask = None

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x seq_length x seq_length]
        head_mask = self.get_head_mask(head_mask, self.config.num_hidden_layers)

        embedding_output = self.embeddings(
            input_ids=input_ids,
            position_ids=position_ids,
            token_type_ids=token_type_ids,
            inputs_embeds=inputs_embeds,
            past_key_values_length=past_key_values_length,
        )
        encoder_outputs = self.encoder(
            embedding_output,
            attention_mask=extended_attention_mask,
            head_mask=head_mask,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=encoder_extended_attention_mask,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        sequence_output = encoder_outputs[0]
        pooled_output = self.pooler(sequence_output) if self.pooler is not None else None

        if not return_dict:
            return (sequence_output, pooled_output) + encoder_outputs[1:]

        return BaseModelOutputWithPoolingAndCrossAttentions(
            last_hidden_state=sequence_output,
            pooler_output=pooled_output,
            past_key_values=encoder_outputs.past_key_values,
            hidden_states=encoder_outputs.hidden_states,
            attentions=encoder_outputs.attentions,
            cross_attentions=encoder_outputs.cross_attentions,
        )

    @classmethod
    def from_pretrained_bert(cls, pretrained_path: str, *args, **kwargs) -> "MemoriaBertModel":
        bert = BertModel.from_pretrained(pretrained_path, *args, **kwargs)
        config = MemoriaBertConfig(**bert.config.to_dict())
        config.model_type = MemoriaBertConfig.model_type
        model = cls(config)
        model.load_state_dict(bert.state_dict(), strict=False)
        return model

    @classmethod
    def from_pretrained_electra(cls, pretrained_path: str, *args, **kwargs) -> "MemoriaBertModel":
        electra = ElectraModel.from_pretrained(pretrained_path, *args, **kwargs)
        config = MemoriaBertConfig(**electra.config.to_dict())
        config.model_type = MemoriaBertConfig.model_type
        model = cls(config)
        model.load_state_dict(electra.state_dict(), strict=False)
        return model


class MemoriaBertForSequenceClassification(MemoriaBertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels
        self.config = config

        self.bert = MemoriaBertModel(config)
        classifier_dropout = (
            config.classifier_dropout if config.classifier_dropout is not None else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple[torch.Tensor], SequenceClassifierOutput]:
        r"""
        labels (`torch.LongTensor` of shape `(batch_size,)`, *optional*):
            Labels for computing the sequence classification/regression loss. Indices should be in `[0, ...,
            config.num_labels - 1]`. If `config.num_labels == 1` a regression loss is computed (Mean-Square loss), If
            `config.num_labels > 1` a classification loss is computed (Cross-Entropy).
        """
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)

        loss = None
        if labels is not None:
            if self.config.problem_type is None:
                if self.num_labels == 1:
                    self.config.problem_type = "regression"
                elif self.num_labels > 1 and (labels.dtype == torch.long or labels.dtype == torch.int):
                    self.config.problem_type = "single_label_classification"
                else:
                    self.config.problem_type = "multi_label_classification"

            if self.config.problem_type == "regression":
                loss_fct = MSELoss()
                if self.num_labels == 1:
                    loss = loss_fct(logits.squeeze(), labels.squeeze())
                else:
                    loss = loss_fct(logits, labels)
            elif self.config.problem_type == "single_label_classification":
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
            elif self.config.problem_type == "multi_label_classification":
                loss_fct = BCEWithLogitsLoss()
                loss = loss_fct(logits, labels)
        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )

    @classmethod
    def from_pretrained_bert(cls, pretrained_path: str, *args, **kwargs) -> "MemoriaBertForSequenceClassification":
        bert = BertForSequenceClassification.from_pretrained(pretrained_path, *args, **kwargs)
        config = MemoriaBertConfig(**bert.config.to_dict())
        config.model_type = MemoriaBertConfig.model_type
        model = cls(config)
        model.load_state_dict(bert.state_dict(), strict=False)
        return model

    @classmethod
    def from_pretrained_electra(cls, pretrained_path: str, *args, **kwargs) -> "MemoriaBertForSequenceClassification":
        electra = ElectraForSequenceClassification.from_pretrained(pretrained_path, *args, **kwargs)
        config = MemoriaBertConfig(**electra.config.to_dict())
        config.model_type = MemoriaBertConfig.model_type
        model = cls(config)
        model.load_state_dict(electra.state_dict(), strict=False)
        return model


MemoriaBertModel.register_for_auto_class("AutoModel")
MemoriaBertForSequenceClassification.register_for_auto_class("AutoModelForSequenceClassification")
AutoModel.register(MemoriaBertConfig, MemoriaBertModel)
AutoModelForSequenceClassification.register(MemoriaBertConfig, MemoriaBertForSequenceClassification)
